unit nbt;

interface
uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs,StdCtrls, ComCtrls, Gauges, uMain, Math, TypInfo;

type
  TNBTType = ( none=0, lc=1, lvl=2);
  PString=^String;
  ByteArray=Array of byte;
  PByteArray=^ByteArray;

NBT_List_Tag = record
  name: string;
  case typ: byte of
    0: ();              //END
    1: (byte:byte);
    2: (short:SmallInt);
    3: (int:Integer);
    4: (long:LongInt);
    5: (float:Single);
    6: (dbl:Double);
    7: (pba:PByteArray);
    8: (str:PString);
end;

  TNBTListArray=Array of NBT_List_Tag;
  PNBTListArray=^TNBTListArray;

NBT_Tag = record
  name: string;
  case typ: byte of
    0: ();              //END
    1: (byte:byte);
    2: (short:SmallInt);
    3: (int:Integer);
    4: (long:Int64);
    5: (float:Single);
    6: (dbl:Double);
    7: (pba:PByteArray);
    8: (str:PString);
    9: (pla:PNBTListArray);
end;

TTagName = (

   TAG_END = 0,       // No name, no payload
   TAG_BYTE = 1,      // char, 8 bits, signed
   TAG_SHORT = 2,     // short, 16 bits, signed
   TAG_INT = 3,       // long, 32 bits, signed
   TAG_LONG = 4,      // long long, 64 bits, signed
   TAG_FLOAT = 5,     // float, 32 bits, signed
   TAG_DOUBLE = 6,    // double, 64 bits, signed
   TAG_BYTE_ARRAY = 7,// char *, 8 bits, unsigned, TAG_INT length
   TAG_STRING = 8,    // char *, 8 bits, signed, TAG_SHORT length
   TAG_LIST = 9,      // X *, X bits, TAG_INT length, no names inside
   TAG_COMPOUND = 10  // nbt_tag *
);


PCompound=^TCompound;
TCompound = record
  name:String;
  tags:Array of NBT_Tag;
  sub:Array of PCompound;
  parent:PCompound;
end;

PTreeView=^TTreeView;

 TNamedBinaryTag = Class(TObject)
 protected
  Cmp:TCompound;
  ptr:PCompound;
  NBTFile:File;
  NBTType:TNBTType;
  bt:Byte;
  s:String;
  PTV:PTreeView;
  procedure Read_Name;
  procedure Read_TAGEnd;
  procedure Read_TAGShort(named:Bool);
  procedure Read_TAGLong(named:Bool);
  procedure Read_TAGInt(named:Bool);
  procedure Read_TAGFloat(named:Bool);
  procedure Read_TAGDouble(named:Bool);
  procedure Read_TAGByteArray(named:Bool);
  procedure Read_TAGString(named:Bool);
  procedure Read_TAGList(named:Bool);
  procedure Read_TAGCompound(named:Bool);
  function  Read_TagByte(named:Bool):byte;

 public
   procedure LoadFromFile(FName:String;G:Tgauge);
   procedure AddToTree(PTV_:PTreeView);


 end;


implementation
{$R *.dfm}

procedure TNamedBinaryTag.Read_Name;
var
  i:Integer;
begin
    BlockRead(NBTFile,bt,1);
    BlockRead(NBTFile,bt,1);
    SetLength(s,bt);
    for I:=1 to bt do
    begin
      BlockRead(NBTFile,bt,1);
      s[i]:=chr(bt);
    end;
end;

procedure TNamedBinaryTag.Read_TAGEnd;
begin
  ptr:=ptr^.parent;
end;
function  TNamedBinaryTag.Read_TagByte(named:Bool):byte;
var
  byte_:Byte;
begin
  if named then
    Read_Name
  else
    s:='';

  if named then
  begin
    BlockRead(NBTFile,byte_,1);
    Result:=byte_;
    ptr^.tags[High(ptr^.tags)].typ := 1;
    ptr^.tags[High(ptr^.tags)].name:=s;
    ptr^.tags[High(ptr^.tags)].byte:=byte_;

    SetLength(ptr^.tags,High(ptr^.tags)+1);
    ptr^.tags[High(ptr^.tags)].typ:=0;


  end
  else
  begin
    BlockRead(NBTFile,byte_,1);
    Result:=byte_;
  end;

end;

procedure TNamedBinaryTag.Read_TAGCompound(named:Bool);
begin
  if named then
  begin
    Read_Name
  else
    s:='';

  SetLength(ptr^.sub,max(Length(ptr^.sub)+1,2));
  ptr^.sub[High(ptr^.sub)]:=nil;
  new(ptr^.sub[High(ptr^.sub)-1]);
  ptr^.sub[High(ptr^.sub)-1].Parent := ptr;
  ptr:=ptr^.sub[High(ptr^.sub)-1];
  ptr^.name:=s;
  SetLength(ptr^.tags,1);
  SetLength(ptr^.sub,1);
  ptr^.sub[0]:=nil;
  ptr^.tags[0].typ:=0;

end;

procedure TNamedBinaryTag.Read_TAGShort(named:Bool);
var
  short:SmallInt;
  b1,b2:Byte;
begin
  if named then
    Read_Name
  else
    s:='';
    BlockRead(NBTFile,b1,1);
    BlockRead(NBTFile,b2,1);
    Short := (b1 SHL 8) + b2;

    ptr^.tags[High(ptr^.tags)].typ := 2;
    ptr^.tags[High(ptr^.tags)].name:=s;
    ptr^.tags[High(ptr^.tags)].short:=Short;

    SetLength(ptr^.tags,Length(ptr^.tags)+1);
    ptr^.tags[High(ptr^.tags)].typ:=0;
end;

procedure TNamedBinaryTag.Read_TAGInt(named:Bool);
var
  int:Integer;
  b1,b2,b3,b4:Byte;
begin
  if named then
    Read_Name
  else
    s:='';

    BlockRead(NBTFile,b1,1);
    BlockRead(NBTFile,b2,1);
    BlockRead(NBTFile,b3,1);
    BlockRead(NBTFile,b4,1);
    int := (b1 SHL 24) + (b2 SHL 16) + (b3 SHL 8) + b4;

    ptr^.tags[High(ptr^.tags)].typ := 3;
    ptr^.tags[High(ptr^.tags)].name:=s;
    ptr^.tags[High(ptr^.tags)].int:=int;

    SetLength(ptr^.tags,Length(ptr^.tags)+1);
    ptr^.tags[High(ptr^.tags)].typ:=0;

  if (NBTType=lvl) then
  begin
    if (s='xPos') then
      uMain.frmMain.LC.xpos := int;
    if (s='zPos') then
      uMain.frmMain.LC.zpos := int;
  end;
end;

procedure TNamedBinaryTag.Read_TAGLong(named:Bool);
var
  long:Int64;
  b1,b2,b3,b4,b5,b6,b7,b8:Byte;
begin
  if named then
    Read_Name
  else
    s:='';

    BlockRead(NBTFile,b1,1);
    BlockRead(NBTFile,b2,1);
    BlockRead(NBTFile,b3,1);
    BlockRead(NBTFile,b4,1);
    BlockRead(NBTFile,b5,1);
    BlockRead(NBTFile,b6,1);
    BlockRead(NBTFile,b7,1);
    BlockRead(NBTFile,b8,1);
    long := b1;
    long := long shl 8;
    long := long + b2;
    long := long shl 8;
    long := long + b3;
    long := long shl 8;
    long := long + b4;
    long := long shl 8;
    long := long + b5;
    long := long shl 8;
    long := long + b6;
    long := long shl 8;
    long := long + b7;
    long := long shl 8;
    long := long + b8;

    ptr^.tags[High(ptr^.tags)].typ := 4;
    ptr^.tags[High(ptr^.tags)].name:=s;
    ptr^.tags[High(ptr^.tags)].long:=long;

    SetLength(ptr^.tags,Length(ptr^.tags)+1);
    ptr^.tags[High(ptr^.tags)].typ:=0;
end;

procedure TNamedBinaryTag.Read_TAGFloat(named:Bool);
var
  float:Single;
  pfloat:^single;
  ba: array[0..3] of byte;
begin
  if named then
    Read_Name
  else
    s:='';

    BlockRead(NBTFile,ba[3],1);
    BlockRead(NBTFile,ba[2],1);
    BlockRead(NBTFile,ba[1],1);
    BlockRead(NBTFile,ba[0],1);
    pfloat:=@ba[0];
    float:=pfloat^;

    ptr^.tags[High(ptr^.tags)].typ := 5;
    ptr^.tags[High(ptr^.tags)].name:=s;
    ptr^.tags[High(ptr^.tags)].float:=float;

    SetLength(ptr^.tags,Length(ptr^.tags)+1);
    ptr^.tags[High(ptr^.tags)].typ:=0;
end;

procedure TNamedBinaryTag.Read_TAGDouble(named:Bool);
var
  dbl:Double;
  pdbl:^Double;
  ba: array[0..7] of byte;
begin
  if named then
    Read_Name
  else
    s:='';

    BlockRead(NBTFile,ba[7],1);
    BlockRead(NBTFile,ba[6],1);
    BlockRead(NBTFile,ba[5],1);
    BlockRead(NBTFile,ba[4],1);
    BlockRead(NBTFile,ba[3],1);
    BlockRead(NBTFile,ba[3],1);
    BlockRead(NBTFile,ba[1],1);
    BlockRead(NBTFile,ba[0],1);
    pdbl:=@ba[0];
    dbl:=pdbl^;

    ptr^.tags[High(ptr^.tags)].typ := 6;
    ptr^.tags[High(ptr^.tags)].name:=s;
    ptr^.tags[High(ptr^.tags)].dbl:=dbl;

    SetLength(ptr^.tags,Length(ptr^.tags)+1);
    ptr^.tags[High(ptr^.tags)].typ:=0;

end;

procedure TNamedBinaryTag.Read_TAGString(named:Bool);
var
  i:Integer;
  len:SmallInt;
  b1,b2:Byte;
  str:String;

begin
  if named then
    Read_Name
  else
    s:='';

    BlockRead(NBTFile,b1,1);
    BlockRead(NBTFile,b2,1);
    len:=(b1 shl 8)+b2;
    SetLength(str,len);
    for I:=1 to len do
    begin
      BlockRead(NBTFile,bt,1);
      str[i]:=chr(bt);
    end;
    ptr^.tags[High(ptr^.tags)].typ := 8;
    ptr^.tags[High(ptr^.tags)].name:=s;
    new(ptr^.tags[High(ptr^.tags)].str);
    ptr^.tags[High(ptr^.tags)].str^:=str;

    SetLength(ptr^.tags,Length(ptr^.tags)+1);
    ptr^.tags[High(ptr^.tags)].typ:=0;
end;

procedure TNamedBinaryTag.Read_TAGByteArray(named:Bool);
var
  i,len:Integer;
  b1,b2,b3,b4:Byte;
  Ba:ByteArray;

begin
  if named then
    Read_Name
  else
    s:='';

  ptr^.tags[High(ptr^.tags)].typ := 7;
  ptr^.tags[High(ptr^.tags)].name:=s;

  BlockRead(NBTFile,b1,1);
  BlockRead(NBTFile,b2,1);
  BlockRead(NBTFile,b3,1);
  BlockRead(NBTFile,b4,1);
  len:= (b1 SHL 24) + (b2 SHL 16) + (b3 SHL 8) + b4;


  SetLength(ba,len+1);
  for i:=0 to len-1 do
    ba[i]:=Read_TAGByte(false);



  new(ptr^.tags[High(ptr^.tags)].pba);
  SetLength(ptr^.tags[High(ptr^.tags)].pba^,len);
  ptr^.tags[High(ptr^.tags)].pba^:=ba;

  SetLength(ptr^.tags,Length(ptr^.tags)+1);
  ptr^.tags[High(ptr^.tags)].typ:=0;
end;
procedure TNamedBinaryTag.Read_TAGList(named:Bool);
var
  i:Integer;
  tagid,b1,b2,b3,b4:Byte;
  len:Integer;
  tPtr:PCompound;
  tCmp:TCompound;
  ListArray:TNBTListArray;
  name:String;
begin
  if named then
    Read_Name
  else
    s:='';

  name:=s;

  BlockRead(NBTFile,bt,1);      //Tags Stored in list
  tagid:=bt;

  BlockRead(NBTFile,b1,1);     //length of list as int
  BlockRead(NBTFile,b2,1);
  BlockRead(NBTFile,b3,1);
  BlockRead(NBTFile,b4,1);
  len:= (b1 SHL 24) + (b2 SHL 16) + (b3 SHL 8) + b4;

if tagid=10 then   //Read  List of Tag Compounds
begin
  for i:=1 to len do     //length read from file
  begin
    Read_TagCompound(false);   //child of list <=> wo. name
    repeat
      BlockRead(NBTFile,b1,1);
      case b1 of
        0 : Read_TAGEnd;
        1 : Read_TAGByte(true);   //child of compound <=> named
        2 : Read_TAGShort(true);
        3 : Read_TAGInt(true);
        4 : Read_TAGLong(true);
        5 : Read_TAGFloat(true);
        6 : Read_TAGDouble(true);
        7 : Read_TAGByteArray(true);
        8 : Read_TAGString(true);
        9 : Read_TAGList(true);
      end;
    until (b1=0);   //Until Tag=END found
  end;
end
else  //read list of anything else to temp compound
begin
  SetLength(tCmp.tags,1);
  tCmp.tags[0].typ:=0;
  tPtr:=ptr;   //Store pointer
  ptr:=@tCmp;  //Set Pointer to local temp compound
  for i:=1 to len do  //length read from file
  begin
    case tagid of
      0 : Read_TAGEnd;
      1 : Read_TAGByte(false);    //child of list <=> wo. name
      2 : Read_TAGShort(false);
      3 : Read_TAGInt(false);
      4 : Read_TAGLong(false);
      5 : Read_TAGFloat(false);
      6 : Read_TAGDouble(false);
      7 : Read_TAGByteArray(False);
      8 : Read_TAGString(false);
      9 : Read_TAGList(False);
      else
      begin
        Showmessage('Error: Unexpected Byte Tag: '+IntTostr(bt)+' in TAG_List');
        break;
      end;
    end;
  end;
  ptr:=tptr; //restore pointer
  ptr^.tags[high(ptr^.tags)].name:='('+IntToStr(len)+') '+name;
  ptr^.tags[high(ptr^.tags)].typ:=9;
  SetLength(ListArray,length(tCmp.tags));
  for i:=0 to high(tCmp.tags) do
  begin
    ListArray[i].name:='';
    ListArray[i].typ:=tagid;
    case tagid of
      1 : ListArray[i].byte := tCmp.tags[i].byte;
      2 : ListArray[i].short := tCmp.tags[i].short;
      3 : ListArray[i].int := tCmp.tags[i].int;
      4 : ListArray[i].long := tCmp.tags[i].long;
      5 : ListArray[i].float := tCmp.tags[i].float;
      6 : ListArray[i].dbl := tCmp.tags[i].dbl;
      7 : ListArray[i].pba := tCmp.tags[i].pba;
      8 : ListArray[i].str := tCmp.tags[i].str;
    end;
  end;
  New(ptr^.tags[High(ptr^.tags)].pla);
  ptr^.tags[High(ptr^.tags)].pla^:=ListArray; 
  SetLength(ptr^.tags,Length(ptr^.tags)+1);
  ptr^.tags[high(ptr^.tags)].typ:=0;
end;

end;

procedure TNamedBinaryTag.LoadFromFile(FName: string;G:TGauge);
begin
  SetLength(cmp.tags,1);
  cmp.tags[0].typ:=0;
  SetLength(cmp.sub,1);
  cmp.sub[0]:=nil;
  ptr:=@cmp;

  NBTType:=TNBTType(0);
  AssignFile(NBTFile,Fname);
  cmp.name:=Fname;
  Reset(NBTFile,1);
  G.MaxValue:=FileSize(NBTFile);
  G.Progress:=0;
   at := GetTickCount; // Anfangszeit
  repeat
  BlockRead(NBTFile,bt,1);
  G.Progress:=FilePos(NBTFile);
//  Showmessage(IntToStr(bt));
  case bt of
  0 : Read_TAGEnd;
  1 : Read_TAGByte(true);
  2 : Read_TAGShort(true);
  3 : Read_TAGInt(true);
  4 : Read_TAGLong(true);
  5 : Read_TAGFloat(true);
  6 : Read_TAGDouble(true);
  7 : Read_TAGByteArray(true);
  8 : Read_TAGString(true);
  9 : Read_TAGList(true);
  10: Read_TAGCompound(true);
  else
  begin
    Showmessage('Error: Unexpected Byte Tag: '+IntTostr(bt));
    break;
  end;
  end;
  until eof(NBTFile);
  et := GetTickCount; // Endzeit
  Showmessage(IntToStr(et-at)); // Zeit et-at in Ticks
  Showmessage(Cmp.name);
end;
procedure TNamedBinaryTag.AddToTree(PTV_: PTreeView);
var
  cPtr:PCompound;
function ListTagValToStr(Tag:NBT_List_Tag):String;
var
  i:Integer;
begin
  case Tag.typ of
    1 : result:=IntToStr(Tag.byte);
    2 : result:=IntToStr(Tag.short);
    3 : result:=IntToStr(Tag.int);
    4 : result:=IntToStr(Tag.long);
    5 : Result:=FloattoStr(Tag.float);
    6 : Result:=FloattoStr(Tag.dbl);
    7 : Result:='['+InttoStr(Length(Tag.pba^)-1)+']';
    8 : Result:=Tag.str^;
  end;
end;
function TagValToStr(Tag:NBT_Tag):String;
var
  i:Integer;
begin
  case Tag.typ of
    1 : result:=IntToStr(Tag.byte);
    2 : result:=IntToStr(Tag.short);
    3 : result:=IntToStr(Tag.int);
    4 : result:=IntToStr(Tag.long);
    5 : Result:=FloattoStr(Tag.float);
    6 : Result:=FloattoStr(Tag.dbl);
    7 : Result:='['+InttoStr(Length(Tag.pba^)-1)+']';
    8 : Result:=Tag.str^;
    9 : begin;
        Result:=' of '+IntToStr(Length(Tag.pla^)-1)+'*';
        Result:=Result+getEnumName(TypeInfo(TTagName),Tag.pla^[0].typ)+': ';
        for i:=0 to High(Tag.pla^)-1 do
          Result:=Result+':'+ListTagValToStr(Tag.pla^[i]);
    end;
  end;
end;
procedure RdTagArray;
var
  i:Integer;
begin
  if cPtr^.tags[0].typ<>0 then
    for i:=0 to High(cPtr^.tags)-1 do
    begin
      PTV^.Items.AddChild(PTV^.Selected,cPtr.tags[i].name +'-'+
                          getEnumName(TypeInfo(TTagName),cptr^.tags[i].typ)+'-'+
                          TagValToStr(cptr^.tags[i])).Selected:=False;
    end;
end;
procedure RdSubCmps;
var
  i:Integer;
begin
  if cPtr.sub[0]<>nil then
  for i:=0 to High(cPtr.sub)-1 do
  begin
    cPtr:=cPtr^.sub[i];
    PTV^.Items.AddChild(PTV.Selected,'Compound: '+cPtr.name).Selected:=True;
    RdTagArray;
    RdSubCmps;
    PTV^.Selected:=PTV^.Selected.Parent;
    cPtr:=cPtr^.parent;
  end;
end;

begin
  PTV:=PTV_;
  PTV^.Items.AddFirst(nil,cmp.name).Selected:=True;
  cPtr:=@cmp;
  RdTagArray;
  RdSubCmps;
end;
end.